import json
import os
from abc import ABCMeta
from modulefinder import Module
from typing import Dict, Optional

import pkg_resources

from fr_emac_gind_rio.driver.rest.RestDriver import RestDriver
from fr_emac_gind_rio.interpretation.abstract_learning_plugin import AbstractLearningPlugin
from fr_emac_gind_rio.models.generated.interpretation_model_api import pjaxb_learnASync, \
    pjaxb_learnASyncResponse, pjaxb_predict, \
    pjaxb_deployResponse, pjaxb_undeployResponse, pjaxb_predictResponse, \
    pjaxb_interpretationConfig, pjaxb_deployResult, pjaxb_applyPredefinedLabelMethod, \
    pjaxb_applyPredefinedLabelMethodResponse
from fr_emac_gind_rio.interpretation.labellisation.abstract_label_method import \
    AbstractLabelMethod
from fr_emac_gind_rio.machine_learning_engine.learn_async_thread import LearnASyncThread
from fr_emac_gind_rio.tensorflow_plugin.progression_tf_callback import ProgessionTFCallback
from fr_emac_gind_rio.utils.regex_helper import RegExHelper
from pkg_resources import EntryPoint

from fr_emac_gind_rio.utils.system_user import DEFAULT_ADMIN


class LearningModuleManager:
    class __LearningModuleManager:

        predefined_labels_methods: Dict[str, AbstractLabelMethod] = {}

        known_plugins: Dict[str, AbstractLearningPlugin] = {}
        deployed_configs_by_ckn: Dict[str, Dict[str, Dict[str, AbstractLearningPlugin]]] = {}

        def __init__(self):
            for entry_point in pkg_resources.iter_entry_points('ml.plugins'):
                ep: EntryPoint = entry_point
                ml_plugin: AbstractLearningPlugin = ep.load().load_plugin()
                self.known_plugins[RegExHelper.to_regex_friendly_name(ep.name).lower()] = ml_plugin

            for entry_point in pkg_resources.iter_entry_points('ml.predefined_label_methods'):
                ep: EntryPoint = entry_point
                module: Module = ep.load()
                predefined_label_method: AbstractLabelMethod = module.__dict__[ep.name]()
                self.predefined_labels_methods[predefined_label_method.get_name()] = predefined_label_method

        def clean(self):
            self.deployed_configs_by_ckn = {}

        def plugin_list(self) -> Dict:
            response: Dict = {
                "module": []
            }
            for key in self.known_plugins.keys():
                module: Dict = {
                    "name": key
                }
                response['module'].append(module)
            return response

        def predefined_labels_methods_list(self) -> Dict:
            response: Dict = {
                "predefined_labels_methods": []
            }
            for key in self.predefined_labels_methods.keys():
                method: AbstractLabelMethod = self.predefined_labels_methods[key]
                method_dict: Dict = {
                    "name": method.get_name(),
                    "description": method.get_description()
                }
                response['predefined_labels_methods'].append(method_dict)
            return response

        def deploy(self, collaboration_name: str, knowledge_space_name: str, plugin_name: str, context: Dict, config: pjaxb_interpretationConfig) -> pjaxb_deployResponse:
            res: pjaxb_deployResponse = pjaxb_deployResponse()
            f_plugin_name: str = RegExHelper.to_regex_friendly_name(plugin_name).lower()
            if self.known_plugins.get(f_plugin_name) is None:
                raise Exception("module not found: " + plugin_name)
            clazz: Optional[ABCMeta] = self.known_plugins.get(f_plugin_name)
            if clazz is not None:
                module: AbstractLearningPlugin = clazz(create_default_config=False, context=context)
                f_collab: str = RegExHelper.to_regex_friendly_name(collaboration_name).lower()
                f_knowledge: str = RegExHelper.to_regex_friendly_name(knowledge_space_name).lower()
                if config is not None and config.context is not None and config.context.resourcesFolderDir is not None:
                    context["resources_folder_base_dir"] = config.context.resourcesFolderDir

                if config is not None:
                    module.load_interpretation_configuration(config, context)
                else:
                    module.load_default_interpretation_configuration(collaborationName=collaboration_name,
                                                                     knowledgeSpaceName=knowledge_space_name,
                                                                     context=context)

                if self.deployed_configs_by_ckn.get(f_collab) is None:
                    self.deployed_configs_by_ckn[f_collab] = {}
                if self.deployed_configs_by_ckn[f_collab].get(f_knowledge) is None:
                    self.deployed_configs_by_ckn[f_collab][f_knowledge] = {}
                self.deployed_configs_by_ckn[f_collab][f_knowledge][f_plugin_name] = module
            res.result = pjaxb_deployResult()
            res.result.add_name(plugin_name)
            return res

        def undeploy(self, collaboration_name: str, knowledge_space_name: str, plugin_name: str) -> pjaxb_undeployResponse:
            f_collab: str = RegExHelper.to_regex_friendly_name(collaboration_name).lower()
            f_knowledge: str = RegExHelper.to_regex_friendly_name(knowledge_space_name).lower()
            f_plugin_name: str = RegExHelper.to_regex_friendly_name(plugin_name).lower()
            del self.deployed_configs_by_ckn[f_collab][f_knowledge][f_plugin_name]
            res: pjaxb_undeployResponse = pjaxb_undeployResponse()
            return res

        def learn_async(self, learn_async: pjaxb_learnASync) -> pjaxb_learnASyncResponse:
            response: pjaxb_learnASyncResponse = pjaxb_learnASyncResponse()



            requestId: str = learn_async.requestId
            collaborationName: str = learn_async.collaborationName
            knowledgeSpaceName: str = learn_async.knowledgeSpaceName
            plugin_name: str = learn_async.pluginName

            context: Dict = self.create_default_context()
            if learn_async.resources_folder_base_dir is not None:
                context["resources_folder_base_dir"] = learn_async.resources_folder_base_dir

            module: AbstractLearningPlugin = self.get_or_deploy_with_default_config(collaborationName, knowledgeSpaceName, plugin_name, context=context)

            thread_learn = LearnASyncThread(module, learn_async, context)
            thread_learn.start()

            #module.learn_async(learn_async, context)
            return response

        def predict(self, pred: pjaxb_predict) -> pjaxb_predictResponse:
            response: pjaxb_predictResponse = pjaxb_predictResponse()


            collaborationName: str = pred.collaborationName
            knowledgeSpaceName: str = pred.knowledgeSpaceName
            plugin_name: str = pred.pluginName

            context: Dict = self.create_default_context()
            if pred.resources_folder_base_dir is not None:
                context["resources_folder_base_dir"] = pred.resources_folder_base_dir

            module: AbstractLearningPlugin = self.get_or_deploy_with_default_config(collaborationName,
                                                                                    knowledgeSpaceName,
                                                                                    plugin_name, context=context)



            driver: RestDriver = RestDriver(context, DEFAULT_ADMIN)
            driver.meta_model_driver.get_usecase_context(knowledge_space_name=knowledgeSpaceName, collaboration_name=collaborationName)
            module.set_driver(driver)


            context["callbacks"] = [ProgessionTFCallback(learningResultByNetworkName=None, context=context, requestId=None, collaborationName=collaborationName, knowledgeSpaceName=knowledgeSpaceName)]

            response = module.predict(pred, context)

            return response

        def get_or_deploy_with_default_config(self, collaboration_name: str, knowledge_space_name: str, plugin_name: str, context: Dict, config: pjaxb_interpretationConfig=None) -> AbstractLearningPlugin :
            module: AbstractLearningPlugin

            isDeployed: bool = True
            f_collab: str = RegExHelper.to_regex_friendly_name(collaboration_name).lower()
            f_knowledge: str = RegExHelper.to_regex_friendly_name(knowledge_space_name).lower()
            f_plugin_name: str = RegExHelper.to_regex_friendly_name(plugin_name).lower()
            if self.deployed_configs_by_ckn.get(f_collab) is None:
                isDeployed = False
                self.deployed_configs_by_ckn[f_collab] = {}
            if self.deployed_configs_by_ckn[f_collab].get(f_knowledge) is None:
                isDeployed = False
                self.deployed_configs_by_ckn[f_collab][f_knowledge] = {}
            if self.deployed_configs_by_ckn[f_collab][f_knowledge].get(f_plugin_name) is None:
                isDeployed = False

            if isDeployed is False:
                self.deploy(collaboration_name=collaboration_name, knowledge_space_name=knowledge_space_name, plugin_name=plugin_name, context=context, config=config)

            module = self.deployed_configs_by_ckn[f_collab][f_knowledge][f_plugin_name]
            if isDeployed is True and config is not None:
                # update interpretation conf
                module.load_interpretation_configuration(conf=config, context=context)
            return module


        def create_default_context(self) -> Dict:
            context: Dict = json.loads(os.environ['rio_config_ml_engine'].replace("'", "\""))

            if context.get("resources_folder_base_dir") is None :
                context["resources_folder_base_dir"] = os.path.dirname(os.path.realpath(__file__)) + "/../../target/resourcesFolder/"

            if not os.path.exists(context["resources_folder_base_dir"]):
                os.makedirs(context["resources_folder_base_dir"])
            return context

        def apply_predefined_label_method(self, req: pjaxb_applyPredefinedLabelMethod) -> pjaxb_applyPredefinedLabelMethodResponse:
            res: pjaxb_applyPredefinedLabelMethodResponse = pjaxb_applyPredefinedLabelMethodResponse();
            method: AbstractLabelMethod = self.predefined_labels_methods[req.methodName]
            res.predefinedLabelMethodReturn = method.create_label_method_return(input_props=req.property)
            return res


    instance = None

    def __new__(c):
        if not LearningModuleManager.instance:
            LearningModuleManager.instance = LearningModuleManager.__LearningModuleManager()
        return LearningModuleManager.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, attr, val):
        return setattr(self.instance, attr, val)


